library("GMMSensitivity")
## Output directory for figures
dir <- "figures/"

## List of different specifications
excluded <- c(6:13, 20:31)
names(excluded) <- c("D/F: \\# cars", "D/F: Sum of horsepower/weight",
                     "D/F: \\# cars with AC", "D/F: Sum Miles/dollar",
                     "D/R: \\# cars", "D/R: Sum of horsepower/weight",
                     "D/R: \\# cars with AC", "D/R: Sum of Miles/dollar",
                     "S/F: \\# cars", "S/F: Sum of log(horsepower/weight)",
                     "S/F: \\# cars with AC", "S/F: Sum of Miles/gallon",
                     "S/F: Sum of log(size)", "S/F: Sum of time trend",
                     "S/R: \\# cars", "S/R: Sum of log(horsepower/weight)",
                     "S/R: \\# cars with AC", "S/R: Sum of Miles/gallon",
                     "S/R: Sum of log(size)", "Supply: Miles/dollar")
ivlist <- c(list(None=vector(mode="numeric", length=0)), as.list(excluded),
            list("All D/F"=6:9, "All D/R"=10:13, "All S/F"=20:25,
                 "All S/R"=26:30, "All excluded demand"=c(6:13),
                 "All excluded supply"=c(20:31), "All excluded"=c(6:13, 20:31)))

## Initial estimates
blp$k_init <- -drop(blp$H %*% solve(crossprod(blp$G, blp$W %*% blp$G),
                                        crossprod(blp$G, blp$W)))
eo <- list(H=blp$H, G=blp$G, Sig=blp$Sig, n=blp$n, g_init=blp$g_init,
           k_init=blp$k_init, h_init= blp$h_init)
B0 <- blp$ZZ %*% diag(sqrt(blp$n)*abs(blp$perturb)/blp$sdZ)

## Columns corresponding to invalid instruments
Ij <- function(j) {
    I <- vector(mode="logical", length=nrow(eo$G))
    I[ivlist[[j]]] <- TRUE
    I
}

eff_ad <- function(j) {
    eb <- function(p)
        EffBounds(eo, B0[, Ij(j), drop=FALSE],
                  switch(as.character(p), "Inf"=1, "2"=sqrt(sum(Ij(j)))), p,
                  beta=0.8, alpha=0.05)
    unlist(c(eb(2), eb(Inf)))
}

js <- c(2, 10, 21:28)
t2 <- t(vapply(js, eff_ad, numeric(4)))
rownames(t2) <- names(ivlist[js])
cat("Table 2")
print(knitr::kable(100*t2[, c(2, 4, 1, 3)], digits=1))

## J test: Here we need to use first-step variance estimate
eoJ <- eo
eoJ$Sig <- solve(blp$W)
Jt <- function(j) {
    c(Jtest(eoJ, B0[, Ij(j), drop=FALSE]*sqrt(sum(Ij(j))), M=1, p=2)$Mmin,
      Jtest(eoJ, B0[, Ij(j), drop=FALSE], M=1, p=Inf)$Mmin)
}

t1 <- t(vapply(js, Jt, numeric(2)))
colnames(t1) <- c("2", "Inf")
rownames(t1) <- names(ivlist[js])
cat("\nTable 1")
print(knitr::kable(t1, digits=2))
cat("J stat:", Jtest(eoJ, B0[, Ij(2), drop=FALSE], M=1, p=2)$J, "\n")


## CIs
js <- c(1, js) # add none
rpoc <- function(p, oc) {
    res <- data.frame()
    for (j in js) {
        Mp <- switch(as.character(p), "Inf"=1, "2"=sqrt(sum(Ij(j))))
        oe <- function(p, oc)
            as.data.frame(OptEstimator(eo, B0[, Ij(j), drop=FALSE], M=Mp, p,
                                       alpha=0.05, opt.criterion=oc)[-1])
        res <- rbind(res, oe(p, oc))
    }
    res
}
rofI <- rpoc(Inf, "FLCI")
rof2 <- rpoc(2, "FLCI")
rin2 <- rpoc(2, "Valid")
rof2$kap <- NULL
rofI$name <- rof2$name <- rin2$name <- names(ivlist[js])

cat("Efficiency gain due to adjusting weight matrix", max(rin2$hl/rof2$hl),
    "\n")


## Figures
## Convert to %
rofI[c(1:4, 6)] <- 100*rofI[c(1:4, 6)]
rof2[c(1:4, 6)] <- 100*rof2[c(1:4, 6)]
rin2[c(1:4, 6)] <- 100*rin2[c(1:4, 6)]

rin2$group <- "Initial"
rof2$group <- "Optimal"

#' @param r A data frame
plotCI <- function(r)  {
    ## So that ggplot doesn't reorder things alphabetically
    r$name <- factor(r$name, levels=unique(r$name))
    r <- droplevels(r)

    p1 <- ggplot2::ggplot(data=r, ggplot2::aes(color=group, shape=group))+
        ggplot2::geom_pointrange(ggplot2::aes(x=name, y=h, ymin=h-hl, ymax=h+hl),
                                 position=ggplot2::position_dodge(width=0.4),
                                 size=0.5) +
        ggplot2::geom_errorbar(ggplot2::aes(x=name, ymin=h-bias, ymax=h+bias),
                               position=ggplot2::position_dodge(width=0.4),
                               size=0.5, width=0.6) +
        ggplot2::labs(y="Average Markup in \\%", x=ggplot2::element_blank(),
                      color="Estimate:",
                      shape="Estimate:") +
        ggplot2::coord_flip()
    p1+ggplot2::theme(axis.text=ggplot2::element_text(size=8),
                      legend.position=c(0.85, 0.15),
                      legend.title=ggplot2::element_text(size=8),
                      legend.text=ggplot2::element_text(size=7))
}

ggthemr::ggthemr(palette = "greyscale", layout = "clean", spacing = 1,
                 text_size=10, type="inner")
f1 <- plotCI(rbind(rin2, rof2))
tikzDevice::tikz(paste0(dir, "blp_l2.tex"), width=5.7, height=4.2,
                 verbose=FALSE)
print(f1)
dev.off()

rof2$group <- "$p=2$"
rofI$group <- "$p=\\infty$"
f2 <- plotCI(rbind(rof2, rofI))
tikzDevice::tikz(paste0(dir, "blp_l2_lI.tex"), width=5.7, height=4.2, ,
                 verbose=FALSE)
print(f2)
dev.off()

## 2. Plot as a function of M ################
I <- vector(mode="logical", length=nrow(eo$G))
I[ivlist[[length(ivlist)]]] <- TRUE
Ms <- seq(0, 4, by=0.01)

spath <- lph(eo, B0[, I], p=Inf)
rpoc <- function(p) {
    res <- data.frame()
    for (M in Ms) {
        Mp <- switch(as.character(p), "Inf"=1, "1"=sum(I), "2"=sqrt(sum(I)))
        oe <- function(p)
            as.data.frame(OptEstimator(eo, B0[, I, drop=FALSE], M=Mp*M, p,
                                       alpha=0.05, opt.criterion="FLCI", spath=spath)[-1])
        res <- rbind(res, oe(p))
    }
    res
}

rof2M <- rpoc(2)
rofIM <- rpoc(Inf)
rof2M$M <- rofIM$M <- Ms
## Convert to %
rofIM[c(1:4, 6)] <- 100*rofIM[c(1:4, 6)]
rof2M[c(1:4, 6)] <- 100*rof2M[c(1:4, 6)]

plotPath <- function(Opt) {
    lM <- nrow(Opt)
    d <- data.frame(x=rep(Opt$M, 3),
                    y=c(Opt$h,
                        Opt$h-Opt$hl, Opt$h+Opt$hl),
                    color=rep(c("E", "B", "B"), each=lM),
                    group=rep(c("E", "Bl", "Bh"), each=lM))
    ggplot2::ggplot(Opt, ggplot2::aes(x=M)) +
        ggplot2::geom_ribbon(ggplot2::aes(, ymin=h-bias, ymax=h+bias),
                             alpha=0.1) +
        ggplot2::geom_line(data=d, ggplot2::aes(x=x, y=y, group=group,
                                                linetype=color)) +
        ggplot2::theme(legend.position="none")+
        ggplot2::labs(x="$M$", y="Average Markup in \\%")
}

f3 <- plotPath(rof2M[rof2M$M <= 2, ])
tikzDevice::tikz(paste0(dir, "blp_l2_M.tex"), width=5.8, height=3.8, verbose=FALSE)
print(f3)
dev.off()
